/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __CVECTOR_HH
	#error "dont include CVector4.hh directly!"
#endif

#ifndef __CVECTOR4_HH
#define __CVECTOR4_HH
/**
 * CHANGELOG:
 *
 * 2008-02-27: martin schreiber
 * 	dist2();
 *
 */

#include <iostream>
#include "CMath.hh"

template <typename T>
class CVector<4,T>
{
public:
	// vector data
	T data[4];

	/*******************
	 * CONSTRUCTURS
	 *******************/
	inline CVector()
	{
		setZero();
	}

	inline CVector(const T x0, const T x1, const T x2, const T x3)
	{
		data[0] = x0;
		data[1] = x1;
		data[2] = x2;
		data[3] = x3;
	}

	inline CVector(const T x)
	{
		data[0] = x;
		data[1] = x;
		data[2] = x;
		data[3] = x;
	}

	inline CVector(const T v[4])
	{
		data[0] = v[0];
		data[1] = v[1];
		data[2] = v[2];
		data[3] = v[3];
	}


	/**
	 * FUNCTINOS
	 */
	inline void setZero()
	{
		data[0] = T(0);
		data[1] = T(0);
		data[2] = T(0);	
		data[3] = T(0);	
	}
#if 0

	inline int createWorldId(const CVector<4,T> &world)
	{
		return data[0] + (data[1] + data[2]*world.data[1])*world.data[0];
	}

	inline long long createWorldIdLongLong(const CVector<3,T> &world)
	{
		return	(long long)data[0] +
			((long long)data[1] + (long long)data[2]*(long long)world.data[1]) * (long long)world.data[0];
	}

	inline CVector<3,T> getNormal()
	{
		CVector<3,T> v = *this;
		float inv_length = 1.0f/sqrtf(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
		v[0] *= inv_length;
		v[1] *= inv_length;
		v[2] *= inv_length;
		return v;
	}

	inline T dot(const CVector<3,T> &v)
	{
		return v.data[0]*data[0] + v.data[1]*data[1] + v.data[2]*data[2];
	}
	inline CVector<3,T> getCrossProd(CVector<3,T> &a)
	{
		return CVector<3,T>(	data[1]*a.data[2] - data[2]*a.data[1],
					data[2]*a.data[0] - data[0]*a.data[2],
					data[0]*a.data[1] - data[1]*a.data[0]
				);
	}

	/**
	 * return square distance to other point
	 */
	inline T dist2(const CVector<3,T> &v)
	{
		CVector<3,T> d = CVector<3,T>(v.data[0] - data[0], v.data[1] - data[1], v.data[2] - data[2]);
		return (d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
	}

	/**
	 * return square distance to other point
	 */
	inline T dist(const CVector<3,T> &v)
	{
		CVector<3,T> d = CVector<3,T>(v.data[0] - data[0], v.data[1] - data[1], v.data[2] - data[2]);
		return sqrtf(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
	}

	/**
	 * return elements in cube
	 * e.g. if the vector components give the resolution of a cube, this function returns the number of cells
	 */
	inline T elements()
	{
		return (data[0]*data[1]*data[2]);
	}

	/**
	 * return length of vector
	 */
#endif
	inline T length()
	{
		return CMath<T>::sqrt(data[0]*data[0] + data[1]*data[1] + data[2]*data[2] + data[3]*data[3]);
	}

	inline T length2()
	{
		return data[0]*data[0] + data[1]*data[1] + data[2]*data[2] + data[3]*data[3];
	}
#if 0
	/**
	 * normalize the vector
	 */
	inline void normalize()	
	{
		T il = 1/length();
		data[0] *= il;
		data[1] *= il;
		data[2] *= il;
	}

	/**
	 * clamp to values -1 and 1
	 */
	inline void clamp1_1()
	{
		data[0] = (data[0] < -1 ? -1 : (data[0] > 1 ? 1 : data[0]));
		data[1] = (data[1] < -1 ? -1 : (data[1] > 1 ? 1 : data[1]));
		data[2] = (data[2] < -1 ? -1 : (data[2] > 1 ? 1 : data[2]));
	}

//	inline void		setVector(const T a[3])				{	data[0] = a[0]; data[1] = a[1]; data[2] = a[2];	};
//	inline void		setVector(const T a0, const T a1, const T a2)	{	data[0] = a0; data[1] = a1; data[2] = a2;	};

	/*******************
	 * ACCESSIBILITY
	 *******************/
#endif
	/*
	inline T operator.x+=(T value)
	{
		data[0] += value;
		return data[0];
	}
	*/


	/*******************
	 * OPERATORS
	 *******************/
	inline CVector<4,T>&	operator=(const T a[4])	{	data[0] = a[0]; data[1] = a[1]; data[2] = a[2];	data[3] = a[3]; return *this;	};
	inline CVector<4,T>&	operator=(CVector<4,T> const & a)	{	data[0] = a.data[0]; data[1] = a.data[1]; data[2] = a.data[2];	data[3] = a.data[3]; return *this;	};
	inline CVector<4,T>&	operator=(CVector<3,T> const & a)	{	data[0] = a.data[0]; data[1] = a.data[1]; data[2] = a.data[2];	data[3] = 1; return *this;	};


// T
	inline CVector<4,T>	operator+(const T a)	{	return CVector<4,T>(data[0]+a, data[1]+a, data[2]+a, data[3]+a);	}
	inline CVector<4,T>	operator-(const T a)	{	return CVector<4,T>(data[0]-a, data[1]-a, data[2]-a, data[3]-a);	}
	inline CVector<4,T>	operator*(const T a)	{	return CVector<4,T>(data[0]*a, data[1]*a, data[2]*a, data[3]*a);	}
	inline CVector<4,T>	operator/(const T a)	{	return CVector<4,T>(data[0]/a, data[1]/a, data[2]/a, data[3]/a);	}
	inline CVector<4,T>&	operator+=(const T a)	{	data[0] += a; data[1] += a; data[2] += a; data[3] += a;	return *this;	}
	inline CVector<4,T>&	operator-=(const T a)	{	data[0] -= a; data[1] -= a; data[2] -= a; data[3] -= a;	return *this;	}
	inline CVector<4,T>&	operator*=(const T a)	{	data[0] *= a; data[1] *= a; data[2] *= a; data[3] *= a;	return *this;	}
	inline CVector<4,T>&	operator/=(const T a)	{	data[0] /= a; data[1] /= a; data[2] /= a; data[3] /= a;	return *this;	}

	inline CVector<4,T>	operator+(const CVector<4,T> &v)	{	return CVector<4,T>(data[0]+v.data[0], data[1]+v.data[1], data[2]+v.data[2], data[3]+v.data[3]);	}
	inline CVector<4,T>	operator-(const CVector<4,T> &v)	{	return CVector<4,T>(data[0]-v.data[0], data[1]-v.data[1], data[2]-v.data[2], data[3]-v.data[3]);	}
	inline CVector<4,T>	operator*(const CVector<4,T> &v)	{	return CVector<4,T>(data[0]*v.data[0], data[1]*v.data[1], data[2]*v.data[2], data[3]*v.data[3]);	}
	inline CVector<4,T>	operator/(const CVector<4,T> &v)	{	return CVector<4,T>(data[0]/v.data[0], data[1]/v.data[1], data[2]/v.data[2], data[3]/v.data[3]);	}

	inline CVector<4,T>&	operator+=(const CVector<4,T> &v)	{	data[0] += v.data[0]; data[1] += v.data[1]; data[2] += v.data[2]; data[3] += v.data[3];	return *this;	}
	inline CVector<4,T>&	operator-=(const CVector<4,T> &v)	{	data[0] -= v.data[0]; data[1] -= v.data[1]; data[2] -= v.data[2]; data[3] -= v.data[3];	return *this;	}

	inline bool	operator==(const CVector<4,T> &v)	{	return bool(data[0] == v.data[0] && data[1] == v.data[1] && data[2] == v.data[2] && data[3] == v.data[3]);	}
	inline bool	operator!=(const CVector<4,T> &v)	{	return bool(data[0] != v.data[0] || data[1] != v.data[1] || data[2] != v.data[2] || data[3] != v.data[3]);	}


	inline T& operator[](const int i)
	{
		return data[i];
	}

	/**
	 * COMPARE SET
	 */
	struct compareSet
	{

		inline bool operator()(CVector<4,float> *v1, CVector<4,float> *v2)
		{
			if ((*v1)[0] != (*v2)[0])
				return (*v1)[0] < (*v2)[0];
			if ((*v1)[1] != (*v2)[1])
				return (*v1)[1] < (*v2)[1];
			if ((*v1)[2] != (*v2)[2])
				return (*v1)[2] < (*v2)[2];
			if ((*v1)[3] != (*v2)[3])
				return (*v1)[3] < (*v2)[3];
			return false;
		}

		inline bool operator()(const CVector<4,T> &v1, const CVector<4,T> &v2)
		{
			if (v1.data[0] != v2.data[0])
				return v1.data[0] < v2.data[0];
			if (v1.data[1] != v2.data[1])
				return v1.data[1] < v2.data[1];
			if (v1.data[2] != v2.data[2])
				return v1.data[2] < v2.data[2];
			if (v1.data[3] != v2.data[3])
				return v1.data[3] < v2.data[3];
			return false;
		}
	};
};


typedef CVector<4,float>	vec4f;
typedef CVector<4,int>	vec4i;


template <class T>
inline
std::ostream&
operator<<(std::ostream &co, CVector<4,T> &v)
{
	return co << "[" << v[0] << ", " << v[1] << ", " << v[2] << ", " << v[3] << "]";
}


#endif
